/* tarot implements the rules of the tarot game
 * Copyright (C) 2019  Vivien Kraus
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "program_commons.h"

static char *default_seed = "default";

static size_t mem = 2 * 1024 * 1024;
static char **seed = &default_seed;
static size_t n_iterations = 100;
static int generate_whole_game = 0;

static void parse_options (int argc, char *argv[]);
static void check_error (TarotMctsError err);

int
main (int argc, char *argv[])
{
  TarotGame *game = NULL;
  char *output;
  TarotMcts *mcts = NULL;
  const TarotGameEvent *event = NULL;
  TarotGameEvent *dog_event;
  size_t n_cards;
  TarotCard *cards;
  tarot_program_setup ();
  parse_options (argc, argv);
  game = tarot_program_read_game ();
  mcts = tarot_mcts_alloc (mem);
  tarot_mcts_seed (mcts, strlen (*seed), *seed);
  while (1)
    {
      if (generate_whole_game)
        {
          TarotPlayer next;
          if (tarot_game_get_next (game, &next) != TAROT_GAME_OK)
            {
              break;
            }
        }
      check_error (tarot_mcts_set_base (mcts, game));
      check_error (tarot_mcts_run (mcts, n_iterations));
      event = tarot_mcts_get_best (mcts);
      if (tarot_game_add_event (game, event) != TAROT_GAME_OK)
        {
          abort ();
        }
      if (tarot_game_can_autoreveal_alloc (game, &n_cards, &cards))
        {
          dog_event = tarot_game_event_alloc_dog (n_cards, cards);
          if (tarot_game_add_event (game, dog_event) != TAROT_GAME_OK)
            {
              abort ();
            }
        }
      if (!generate_whole_game)
        {
          break;
        }
    }
  output = tarot_game_save_to_xml_alloc (game);
  printf ("%s", output);
  return EXIT_SUCCESS;
}

static void
parse_options (int argc, char *argv[])
{
  int c;
  char *end;
  while (1)
    {
      static struct option long_options[] = {
        {N_("help"), no_argument, 0, 'h'},
        {N_("version"), no_argument, 0, 'v'},
        {N_("max-memory"), required_argument, 0, 'm'},
        {N_("seed"), required_argument, 0, 's'},
        {N_("n-iterations"), required_argument, 0, 'i'},
        {N_("mcts-parameter"), required_argument, 0, 'p'},
        {N_("parameter-random"), required_argument, 0, 'r'},
        {N_("parameter-agreed"), required_argument, 0, 'a'},
        {N_("generate"), no_argument, 0, 'g'},
        {0, 0, 0, 0}
      };
      int option_index = 0;
      for (option_index = 0; long_options[option_index].name != NULL;
           option_index++)
        {
          long_options[option_index].name =
            _(long_options[option_index].name);
        }
      option_index = 0;
      c = getopt_long (argc, argv, "hvm:s:i:g", long_options, &option_index);
      if (c == -1)
        break;
      switch (c)
        {
        case 0:
          break;

        case 'h':
          printf (_("Usage: tarot-mcts [OPTIONS]... < GAME\n\
Play as the MCTS in GAME.  Options:\n\
- -h, --help: print this message and exit;\n\
- -v, --version: print the program version and exit;\n\
- -m SZ, --max-memory=SZ: define the maximum memory (default: %lu B).\n\
   You can use the following suffixes: 'GiB', 'MiB', 'KiB', or 'B'.\n\
   By default, numbers will be treated as bytes;\n\
- -s SEED, --seed=\"SEED\": set the seed for the random deal.  SEED\n\
   must not be empty;\n\
- -i N, --n-iterations=N: set the number of iterations.  The higher\n\
   the better.  Default: %lu;\n\
- -g, --generate: play with the same player parameters until the end\n\
   of the game.\n\
\n\
If the number of iterations is too low, or the algorithm is extremely\n\
unlucky, then the operation will fail.\n\
"), mem, n_iterations);
          exit (EXIT_SUCCESS);
          break;

        case 'v':
          printf (_("%s (libtarot %s)\n\
\n\
Copyright © 2019 Vivien Kraus\nThis is free software; see the source for copying conditions.  There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"), "tarot-mcts", tarot_libtarot_version ());
          exit (EXIT_SUCCESS);
          break;

        case 'm':
          mem = tarot_strtosz (optarg, &end);
          if (strcmp (end, optarg) == 0)
            {
              fprintf (stderr,
                       _
                       ("Error: the memory limit should be a size in bytes, not '%s'.\n"),
                       optarg);
              exit (EXIT_FAILURE);
            }
          else if (strcmp (end, "") != 0)
            {
              fprintf (stderr,
                       _
                       ("Error: the memory limit (%lu bytes) has an unexpected suffix, '%s': '%s'.\n"),
                       mem, end, optarg);
              exit (EXIT_FAILURE);
            }
          break;

        case 's':
          *seed = strdup (optarg);
          if (strlen (*seed) == 0)
            {
              fprintf (stderr,
                       _
                       ("Error: the seed must not be empty.  The default is usually \"default\".\n"));
              exit (EXIT_FAILURE);
            }
          break;

        case 'i':
          n_iterations = strtoul (optarg, &end, 10);
          if (strcmp (end, optarg) == 0)
            {
              fprintf (stderr,
                       _
                       ("Error: the number of iterations should be a number, not '%s'.\n"),
                       optarg);
              exit (EXIT_FAILURE);
            }
          else if (strcmp (end, "") != 0)
            {
              fprintf (stderr,
                       _
                       ("Error: the number of iterations (%lu) has an unexpected suffix, '%s': '%s'.\n"),
                       n_iterations, end, optarg);
              exit (EXIT_FAILURE);
            }
          break;

        case 'g':
          generate_whole_game = 1;
          break;

        case '?':
          /* getopt_long already printed an error message. */
          exit (EXIT_FAILURE);
          break;

        default:
          abort ();
        }
    }
  if (optind < argc)
    {
      fprintf (stderr, _("Error: extra arguments:\n"));
      while (optind < argc)
        fprintf (stderr, _("- %s\n"), argv[optind++]);
      exit (EXIT_FAILURE);
    }
}

static void
check_error (TarotMctsError err)
{
  switch (err)
    {
    case TAROT_MCTS_NONEXT:
      fprintf (stderr, _("Error: no player needs to play in this game.\n"));
      exit (EXIT_FAILURE);
    case TAROT_MCTS_UNKNOWNNEXT:
      fprintf (stderr,
               _("Error: we don't know the cards of the next player.\n"));
      exit (EXIT_FAILURE);
    case TAROT_MCTS_NODATA:
      fprintf (stderr,
               _
               ("Error: too few experiments.  Please use more iterations.\n"));
      exit (EXIT_FAILURE);
    case TAROT_MCTS_NOMEM:
      fprintf (stderr,
               _
               ("Error: memory exceeded.  Please pass --max-memory=SIZE.\n"));
      exit (EXIT_FAILURE);
    case TAROT_MCTS_OK:
      break;
    }
}
